{{
        IntADC - Integrating Sigma-Delta ADC on the Propeller
        
        Jonathan "lonesock" Dummer

        Using a capacitor in series with a signal blocks DC,
        but is basically a differentiator, where high frequencies
        come through stronger than low frequencies.  This code
        numerically integrates the ADC input (keeping it centered
        around 0), returning the integrated signal into a Hub long.

        v-v-v Below is the original explanation for the raw ADC part v-v-v

        Sigma-Delta ADC is very easy with the propeller.  This object shows a simple
        implementation of a 2-channel ADC input.  Here's how to wire it up:

        Simplest ADC circuit:

        In ----------+ <= common node, at threshold voltage
                     |
        Out ---Rout--+---Rmeasure---* Voltage_to_measure
                     |
        Gnd ----C----+

        
        Using counter mode POS (or NEG) w/ feedback, the propeller makes Out-pin the opposite of In-pin,
        while counting the number of clock cycles that A was high (or low for NEG mode).  The threshold
        for the input pin is halfway between VDD and Gnd, or ~1.6V. If the common node voltage is above the
        threshold, the In pin will read high, and the propeller's counter will drive the Out pin Low, and
        vice-versa, effectively pulling the common node to the threshold voltage.  Over a given time
        period, the percentage of time spent pulling the common node voltage low is proportional to the
        input Voltage.

        The capacitor acts in conjunction with the resistors to act as a RC low-pass filter for the Out-
        pin's output.  The capacitor is needed, otherwise as soon as the Out pin is set, the voltage at
        the common node changes instantaneously, with no chance for the counter to actually count anything.
        So, the cap and resistor are a simple RC filter, smoothing out high-frequency noise, with the
        corner-frequency of the filter at 1/Tau, where Tau is the time constant of the filter, which for
        a RC filter is simply:

                Tau = Rout*C

        Various values of C seem to work fine, so I am recommending setting the C value as if you were using
        the RC filter for a DAC, so Rout*C ~= 1 / typical_output_sample_frequency.  NOTE: this recommendation
        is subject to change!!

        With a little bit of Kirchhoff's Current Law (KCL, a.k.a. "summing the current into a node should equal 0"),
        and assuming steady state (so the capicitors act as an open circuit):
        * The maximum input is hit when the Out pin is driven low 100% of the time: Voltage = 1.6 * (1 + Rmeasure / Rout)
        * The minimum input is hit when the Out pin is driven high 100% of the time: Voltage = 1.6 * (1 - Rmeasure / Rout)

        The maximum value coming out of the counter is simply the period over which you sample.  For example,
        with a 80 MHz system, running a sampling rate of 44.1 kHz means the sampling period is 1814 clocks.
        The minimum value coming out of the counter is obviously 0, and the maximum possible value of the counter
        is 1814 (if the In pin was high the whole time, the counter would add 1 to PHSx for each clock).  This
        is equivalent to about 10.8 bits.  If you wanted the module to report a 16-bit number exactly, you
        would need to make sure the sampling period was 65535 (or 1.22 kHz on a 80 MHz system).  To make the
        output of the system easier to use, I adjust the FRQx value so that period * FRQx = 2^31-1.  That way,
        after a full period is complete, I can shift right by 15 bits, yielding a 16-bit number.  Note that
        this does _NOT_ mean you are getting 16 significant bits, only that the reading are always scaled to
        that range. 
        
        Things to note:

        * you want large resistor values if you are trying to measure an unbuffered voltage source
                - for example, if you want to measure the voltage on a capacitor...the circuit
                  will bleed current (V_measure_me - 1.6) / Rmeasure.  If you are measuring the
                  output of an op-amp, the resistor values you choose are no big deal (other than
                  setting gain).  Note that the larger the Resistors, the noiser the system can get.
        * keep your traces as short as possible
        * use non-adjascent pins for In and Out if possible (to reduce cross-talk)
        * if possible use a low cog for pins P0..P15 (cog 0 is best, then 1, etc)
        * if possible use a high cog for pins P16..P31 (cog 7 is best, then 6, etc.)
        * there is no point summing then averaging N values...it's the same as using sample-rate / N
        * averaging is still useful if you are smoothing a waveform
        * restart the cog to change the sampling rate
        * the PASM loop takes about 80 clocks, so don't try to go over clkfreq/80 Hz for the sample-rate

        
        Here is a more complex version of the circuit, to add some features:

             +-Rbias-+ 
             |       |
        VDD -+---C---+ <= common node 
                     |
        In -----Rin--+---Rmeasure--- Voltage to measure
                     |
        Out ----Rout-+
                     |
        Gnd -----C---+

        Rin does nothing as the input pin is high impedence, but can be used to switch the function
        of the In and Out pins, changing the ADC's range.  For example, if Rin = 0.5*Rout, switching pin
        functions (Out is now the sense pin, and In is now the feedback pin) would yield a new range that
        is 2x larger.

        Note that if you have 2 capacitors to constant voltages (VDD and GND), the equivalent capacitance
        of the system looks like both capacitors are in parallel, or 2*C (2 caps are very useful since
        noise on either VDD or Gnd can affect the threshold voltage).

        Rbias can be left out of the circuit as well, but since the sum of the current coming into the
        node is 0, you can use Rbias to inject a constant current into the node (since VDD is constant,
        and the threshold voltage is constant (thanks to the prop's counter) we are basically offsetting
        the center point).  As a useful example, if Rbias == Rmeasure, you effectively shifted the input
        range to be +- 1.6 * Rmeasure/Rout, centered around 0 V!  You can verify this using KCL.  The more
        intuitive way to see this, at least for me, is to note that Rmeasure & Rbias form a resistive divider
        between Vin and VDD such that the common node is exactly 1/2 way between the the two, i.e. when the
        input voltage is 0 then the common node voltage would be at VDD/2, which is already the threshold
        voltage, and the counter doesn't have to sink or source any current.  If you want to push the center
        point around you can change Rbias, and/or connect it to Gnd instead of VDD.  For extreme flexibility,
        you could even use a potentiometer with one leg each to Gnd and VDD, or even another output pin via
        a RC filter DAC connected via a resistor to the common node, to affect fine tuning electronically.         
}}

CON
  leaky_bits    = 5             ' keep the audio centered around 0!

VAR
  byte cog

PUB start_freq( Ain, Aout, SampleHz, ptr )
  '' helper function, to convert Sample Hertz to (rounded) period in system clocks
  return start_period( Ain, Aout, (clkfreq + (SampleHz >> 1)) / SampleHz, ptr )  

PUB start_period( Ain, Aout, PeriodClocks, ptr )
  '' initialize the sigma-delta ADC cog
  stop
  drive_mask := (|<Aout)
  period := PeriodClocks #> 150 ' takes a little less than 145 clocks per loop
  ctra_val := constant(%01001 << 26) | (Aout << 9) | (Ain)
  frq_val := POSX / period ' scale to 1<<31 - 1 (all bits 1 except the top bit)
  mid_input := frq_val * period / 2 
  return cog := cognew( @IntADC_pasm, ptr ) + 1

PUB stop
  '' kill the cog
  if cog
    cogstop( cog~ - 1 )

DAT
ORG 0

IntADC_pasm   ' Initialize hardware
              mov       ctra, ctra_val
              mov       frqa, frq_val
              mov       dira, drive_mask

              ' Initialize variables
              mov       timeout, period
              add       timeout, cnt
              mov       phsa, last_phsa

              
main_loop     '' LOOP FOREVER

              ' grab the latest sample (will start unsigned, in 31 bits)
              waitcnt   timeout, period         ' precise timing FTW!
              mov       A, phsa                 ' grab the counter phase value
              sub       A, last_phsa            ' subtract previous to get the change
              add       last_phsa, A            ' update the tracker
              sub       A, mid_input            ' center around 0; faster settling, and NEEDED if not removing DC bias 

              '' Simple IIR Low-Pass-Filter for A (signed)
              add       lastA, A
              sar       lastA, #1
              mov       A, lastA                 
              
              '' Remove DC bias from the signal
              add       dc_accumA, dc_valueA    ' find a fraction to remove (leak)
              mov       tmp, dc_accumA          ' add any residual error from previous leaks
              and       dc_accumA, dc_mask      ' well, only the portions that will not affect this round
              sar       tmp, #leaky_bits        ' do the scaling
              sub       dc_valueA, tmp          ' do the leaking                
              sub       A, track_A              ' straight delta blocks all DC
              add       track_A, A              ' update the trailing value
              add       dc_valueA, A            ' add in the new delta 
              mov       A, dc_valueA            ' copy it back into my temporary variable

              '' Remove DC bias again, this time for the integrated value: B
              add       dc_accumB, dc_valueB    ' find a fraction to remove (leak)
              mov       tmp, dc_accumB          ' add any residual error from previous leaks
              and       dc_accumB, dc_mask      ' well, only the portions that will not affect this round
              sar       tmp, #leaky_bits        ' do the scaling
              sub       dc_valueB, tmp          ' do the leaking
              add       dc_valueB, A            ' add in the new delta, A from the ADC
              mov       B, dc_valueB            ' copy it back into my temporary variable

              '' Simple IIR Low-Pass-Filter for B (signed)
              add       lastB, B
              sar       lastB, #1
              mov       B, lastB
                
              '' Store it
              wrlong    B, par                  { integrated value
              ' OR
              shl       A, #2                   ' this is quiter...boosted for comparison
              wrlong    A, par                  ' vanilla           }
              jmp       #main_loop
              
              
                                                       

{===== PASM Parameters and 'constants' =====}
drive_mask    long      0
ctra_val      long      0
ctrb_val      long      0
period        long      0
frq_val       long      0
limitFFFF     long      $FFFF                                                                 

A             long      0
B             long      0
lastA         long      0
lastB         long      0
track_A       long      0
dc_accumA     long      0
dc_accumB     long      0
dc_valueA     long      0
dc_valueB     long      0

sync_history  long      0
mid_input     long      POSX / 2

dc_mask       long      |<leaky_bits - 1

debug_val     long      0
debug_ptr     long      0


{===== PASM scratch variables =====}
timeout       res       1
buf_ptr       res       1
buf_cnt       res       1
filter0       res       1
filter1       res       1
oldA1         res       1    
oldA2         res       1      
oldB1         res       1      
oldB2         res       1
last_phsa     res       1
last_phsb     res       1
tmp           res       1

cnt_tot       res       1
cnt_used      res       1